"""
Created on Sat Mar  4 20:42:14 2017

Extract a date and or time from a string in a variety of formats

@author: David
"""

"""
Possible formats for date:
    Dates:
    25/04/94,25/04/1994, 25/04, 25th Apr, Apr 25th, April 25th
    The 25th of this month, The 25th of next month, 1 month today, 2 months today, in two weeks
    a week on saturday, tomorrow, yesterday, in two days, on Friday ...

    How to approach this problem?
    1) Parse the text, splitting into sections seperated by wither spaces or symboles
    2) Run through the text systematically analysing each 'word' and it's significance
    3) Return the best possible understanding
"""
from datetime import datetime, timedelta
import calendar
from dateutil import rrule

from . import num_parse
from .dates import date_finder, time_finder

# import num_parse
# from dates import date_finder, time_finder

from textblob import TextBlob

#A dictionary of words that can be understood for mapping words of the same meaning and grouping
_date_words = {
        'second':['time','sec'],
        'seconds':['time','sec'],
        'sec':['time','sec'],
        'secs':['time','sec'],
        'minute':['time','min'],
        'minutes':['time','min'],
        'min':['time','min'],
        'mins':['time','min'],
        'hour':['time','hr'],
        'hours':['time','hr'],
        'hr':['time','hr'],
        'hrs':['time','hr'],
        'day':['time','day'],
        'days':['time','day'],
        'week':['time','wk'],
        'weeks':['time','wk'],
        'wk':['time','wk'],
        'month':['time','mth'],
        'months':['time','mth'],
        'mth':['time','mth'],
        'mths':['time','mth'],
        'year':['time','yr'],
        'yr':['time','yr'],
        'years':['time','yr'],
        'yrs':['time','yr'],
        'morning':['time','am'],
        'afternoon':['time','pm'],
        'evening':['time','pm'],
        'night':['time','pm'],
        'am':['time','am'],
        'pm':['time','pm'],
        'monday':['day',0],
        'mon':['day',0],
        'tuesday':['day',1],
        'tue':['day',1],
        'wednesday':['day',2],
        'wed':['day',2],
        'thursday':['day',3],
        'thur':['day',3],
        'thu':['day',3],
        'friday':['day',4],
        'fri':['day',4],
        'saturday':['day',5],
        'sat':['day',5],
        'sunday':['day',6],
        'sun':['day',6],
        'weekend':['day','wk_end'],
        'weekends':['day','wk_end'],
        'january':['mth',1],
        'jan':['mth',1],
        'february':['mth',2],
        'feb':['mth',2],
        'march':['mth',3],
        'mar':['mth',3],
        'april':['mth',4],
        'apr':['mth',4],
        'apl':['mth',4],
        'may':['mth',5],
        'june':['mth',6],
        'jun':['mth',6],
        'july':['mth',7],
        'jul':['mth',7],
        'august':['mth',8],
        'aug':['mth',8],
        'september':['mth',9],
        'sept':['mth',9],
        'sep':['mth',9],
        'october':['mth',10],
        'oct':['mth',10],
        'november':['mth',11],
        'nov':['mth',11],
        'december':['mth',12],
        'dec':['mth',12],
        'yesterday':['rel','yest'],
        'today':['rel','td'],
        'tomorrow':['rel','tmrw'],
        'in':['rel','in'],
        'last':['rel','last'],
        'first':['rel','first'],
        'this':['rel','this'],
        'next':['rel','nxt'],
        'every':['rel','evy'],
        'at':['at','at'],
        'on':['on','on'],
        'for':['len','for'],
        'untill':['len','till'],
        'till':['len','till'],
        'a':['a','a'],
        'an':['a','an'],
        'midnight':['num','12am'],
        'midday':['num','12pm'],
        'noon':['num','12pm'],
        'past':['num','past'],
        'to':['num','to'],
        'half':['num','half'],
        'quarter':['num','qtr'],
        }

class natural_time_parser():

    def __init__(self):
        # Variables for storing the date
        self.now = datetime.now()
        self.year = 0
        self.month = 0
        self.day = 0
        self.hour = 0
        self.minute = 0
        self.second = 0
        self.length = 0
        self.morning = None
        # Perameters that will be passed to rrule
        self.freq = rrule.MINUTELY
        # Frequency can be either rrule.YEARLY, MONTHLY, WEEKLY, DAILY, HOURLY, MINUTELY or SECONDLY
        self.dtstart = datetime.now()  # Start from today
        self.interval = 1
        self.wkstart = None
        self.count = None  # Returns all instances as default
        self.until = self.dtstart + timedelta(days=365*4+1)
        self.bysetpos = None
        self.byyear = None
        self.bymonth = None
        self.bymonthday = None
        self.byyearday = None
        self.byweekno = None
        self.byweekday = None # 0 = Monday ... 6 = Sunday
        self.byhour = None
        self.byminute = None
        self.bysecond = 0 # Default to 0 (seconds)

        # self.log = Path(os.path.dirname(__file__))/'date_log.txt'

        self._tags = {
                'time',
                'day',
                'mth',
                'rel',
                'at',
                'len',
                'num',
                'rel_t',
                'tm',
                'date',
                'at',
                'one',
                'a',
                }



    def parse_string(self,s):
        # First make the string all lower case, then split up all the words
        s = s.lower()
        # Convert all numerical text to numbers (one => 1, second => 2nd etc
        s = num_parse.convert_num(s)
        tb = TextBlob(s)
        words = tb.words
        # Map and group the words using _date_words
        date_mapped = [(_date_words[w] if w in _date_words else w) for w in words]
        # Remove all non-important words and tag numbers
        date_tagged = []
        for item in date_mapped:
            if self.contains_digit(item):
                date_tagged.append(['num', item])
            elif item[0] in self._tags:
                date_tagged.append(item)
        # Cycle through everything and try to interpret the meaning of all the numbers
        self.interpret_num(date_tagged)
        # Now the numbers are turned to proper numbers and the words are tagged try to understand
        self.interpret_tags(date_tagged)

        # Check the bymonthday value and bymonth value are compatible
        if self.bymonth != None and self.bymonthday != None:
            # Which year are we probably looking at
            if self.byyear == None:
                if self.now.month > self.bymonth or (self.now.month == self.bymonth and self.bymonthday < self.now.day):
                    self.byyear = 2018
                else:
                    self.byyear = 2017
            max_days = calendar.monthrange(self.byyear,self.bymonth)[1]
            if self.bymonthday > max_days and self.bymonthday != 29:
                return []
        elif self.bymonthday != None and self.bymonthday > 31:
            return []

        self.until = None
        self.count = 1

        date_list = list(rrule.rrule(
                freq=self.freq, dtstart=self.dtstart, interval=self.interval,
                wkst=self.wkstart, count=self.count, until=self.until,
                bysetpos=self.bysetpos, bymonth=self.bymonth,
                bymonthday=self.bymonthday, byyearday=self.byyearday,
                byweekno=self.byweekno, byweekday=self.byweekday,
                byhour=self.byhour, byminute=self.byminute,
                bysecond=self.bysecond))

        date = date_list[0]

        return date

    def contains_digit(self, st):
        if isinstance(st, str):
            try:
                return any(c.isdigit() for c in st)
            except Exception:
                return False
        else:
            return False

    def interpret_num(self, st_tagged):
        '''
        Seeks to understand the meening of each item tagged with the 'num' tag.
        Is it a day? Month? Time? Date? etc
        '''
        i = 0
        for item in st_tagged:
            if item[0] != 'num':
                i += 1
                continue
            # Special case where 'past' and 'to' have been found
            if item[1] == 'past':
                # Check for number before and after
                if st_tagged[i-1][0] == 'num' and st_tagged[i+1][0] == 'num':
                    # Check for half and qtr
                    if st_tagged[i-1][1] == 'half':
                        item = ['num', st_tagged[i+1][1]+':'+'30']
                    elif st_tagged[i-1][1] == 'qtr':
                        item = ['num', st_tagged[i+1][1]+':'+'15']
                    else:
                        item = ['num', st_tagged[i+1][1]+':'+st_tagged[i-1][1]]
            elif item[1] == 'to':
                if st_tagged[i-1][0] == 'num' and st_tagged[i+1][0] == 'num':
                    # This puts the time in a format that will be recognized
                    if st_tagged[i-1][1] == 'half':
                        minutes = '30'
                    elif st_tagged[i-1][1] == 'qtr':
                        minutes = '45'
                    else:
                        minutes = str(60 - int(st_tagged[i-1][1]))
                    hour = str(int(st_tagged[i+1][1]) - 1)
                    item = ['num',hour+':'+minutes]
            elif item[1] == 'half' and st_tagged[i+1][0] == 'num':
                hour = str(st_tagged[i+1][1])
                minutes = '30'
                item = ['num',hour+':'+minutes]

            # Special cases where the number is a year (previous word a month and/or > 1000) or day ( <=32 next to a month) or hour (next to am/pm)
            try:
                num = int(item[1])
                if num > 1000 and (st_tagged[i-1][0] == 'mth' or st_tagged[i+1][0] == 'mth'):
                    date = ['date',[None,None,num]]
                    st_tagged[i] = date
                    i += 1
                    continue
                elif num <= 32 and (st_tagged[i-1][0] == 'mth' or st_tagged[i+1][0] == 'mth'):
                    date = ['date',[num,None,None]]
                    st_tagged[i] = date
                    i += 1
                elif num <= 23 and (st_tagged[i+1][1] == 'am' or st_tagged[i+1][1] == 'pm'):
                    item = ['num',str(num)+':00']
            except Exception:
                pass
            # TODO Special case where the number 1st, 2nd, 3rd etc to understand different meanings

            # Check if the number is some form of time or date.
            time = time_finder(item[1])
            date = date_finder(item[1])
            # Interpret the time or date
            if time != None:
                st_tagged[i] = time
                i += 1
                continue
            elif date != None:
                st_tagged[i] = date
                i += 1
                continue
            i += 1

    def interpret_tags(self,st_tagged):
        '''
        Tries to understand the meaning of a the phrase in relation to dates and time.
        Tags are interpreted in the following order:
            rel, len, time, day, mth, num
        '''
        # Work through each word, analysing the tag and meaning.
        # Note that once anything has been interpreted (even if by another function) it is not looked at again
        for i, w in enumerate(st_tagged):
            if w[0] == 'rel':
                self.rel_tag(st_tagged,i)

        for i, w in enumerate(st_tagged):
            if w[0] == 'len':
                self.len_tag(st_tagged,i)

        for i, w in enumerate(st_tagged):
            if w[0] == 'time':
                self.time_tag(st_tagged,i)

        for i, w in enumerate(st_tagged):
            if w[0] == 'day':
                self.day_tag(st_tagged,i)

        for i, w in enumerate(st_tagged):
            if w[0] == 'mth':
                self.mth_tag(st_tagged,i)

        for i, w in enumerate(st_tagged):
            if w[0] == 'at':
                self.at_tag(st_tagged,i)

        for i, w in enumerate(st_tagged):
            if w[0] == 'num':
                self.num_tag(st_tagged,i)

        for i, w in enumerate(st_tagged):
            if w[0] == 'tm':
                self.tm_tag(st_tagged,i)

        for i, w in enumerate(st_tagged):
            if w[0] == 'rel_t':
                self.rel_t_tag(st_tagged,i)

        for i, w in enumerate(st_tagged):
            if w[0] == 'date':
                self.date_tag(st_tagged,i)

    def date_tag(self,st_tagged,i):
        '''Pretty straight forward, it's a date'''
        self.bymonthday = st_tagged[i][1][0] if st_tagged[i][1][0] != None else self.bymonthday
        self.bymonth = st_tagged[i][1][1] if st_tagged[i][1][1] != None else self.bymonth
        #For the year need to change the start date
        if st_tagged[i][1][2] != None:
            new_year = st_tagged[i][1][2]
            if self.now.year < new_year:
                self.byyear = new_year
                self.dtstart = self.dtstart.replace(year=new_year,month=1,day=1,hour=0,minute=0,second=0)
                self.until = self.dtstart.replace(month=12,day=31,hour=23,minute=59,second=59)
                #TODO feedback when event is in the past
        st_tagged[i] = [None,None]

    def at_tag(self,st_tagged,i):
        '''
        Could be followed by lots of things, generally denots a time or place.
        In this case time is assumed so we check the following 'word' for a number
        or time.
        '''
        try:
            number = int(st_tagged[i+1][1])
            if number < 24:
                time = ['tm',[number,0,0,None]]
            st_tagged[i+1] = time
        except (TypeError, IndexError, ValueError):
            pass
        st_tagged[i] = [None,None]


    def rel_t_tag(self,st_tagged,i):
        '''
        Finds the future moment represented by the relative time
        '''
        time = st_tagged[i][1]
        future = self.now + timedelta(hours=time[0],minutes=time[1],seconds=time[2])
        self.start = future.replace(hour=0,minute=0,second=0)
        self.bymonthday = future.day
        self.bymonth = future.month
        self.byhour = future.hour
        self.byminute = future.minute
        self.bysecond = future.second
        st_tagged[i] = [None,None]

    def tm_tag(self,st_tagged,i):
        '''
        It is assumed currently that this is a start time
        TODO fix for situations where this is an end time e.g 10pm to 11pm
        '''
        self.byhour = st_tagged[i][1][0]
        self.byminute = st_tagged[i][1][1]
        self.bysecond = st_tagged[i][1][2]
        self.morning = st_tagged[i][1][3] if st_tagged[i][1][3] != None else self.morning
        #Check that it matches correct AM/PM
        if self.morning == False and self.byhour < 12:
                self.byhour += 12
        elif self.morning == True and self.byhour >= 12:
            self.byhour -= 12
        elif self.morning == None and self.byhour <= 6:
            self.byhour += 12
        st_tagged[i] = [None,None]

        return

    def time_tag(self,st_tagged,i):
        '''
        Interprets items tagged with the 'time' tag.
        Words included in the tag are: second (sec), minute (min), hour (hr)
        day (day), week (wk), month (mth), year (yr), am and pm (am, pm).
        TODO get to work with the different time tags (sec, mth, wk etc)
        '''
        #First check for am or pm
        try:
            new_tag = [None, None]
            if st_tagged[i][1] == 'am':
                self.morning = True
            elif st_tagged[i][1] == 'pm':
                self.morning = False
            elif st_tagged[i][1] == 'wk':
                #Look at the following words to check for phrase such as 'week on sunday'
                try:
                    if st_tagged[i+1][0] == 'day':
                        date = (self.now + timedelta(days=8)).replace(hour=0,minute=0,second=0)
                        self.dtstart = date
                        self.byweekday = st_tagged[i+1][1]
                        self.until = (date + timedelta(days=6)).replace(hour=23,minute=59,second=59)
                except IndexError:
                    pass
            if st_tagged[i-1][0] == 'num':
                num = int(st_tagged[i-1][1])
                if st_tagged[i][1] == 'sec':
                    new_tag = ['rel_t', [0, 0, num]]
                elif st_tagged[i][1] == 'min':
                    new_tag = ['rel_t', [0, num, 0]]
                elif st_tagged[i][1] == 'hr':
                    new_tag = ['rel_t', [num, 0, 0]]
                elif st_tagged[i][1] == 'day':
                    new_tag = ['rel_t', [num*24, 0, 0]]
                elif st_tagged[i][1] == 'wk':
                    new_tag = ['rel_t', [num*24*7, 0, 0]]
                elif st_tagged[i][1] == 'mth':
                    new_tag = ['rel_t', [num*24*31, 0, 0]]
                elif st_tagged[i][1] == 'yr':
                    new_tag = ['rel_t', [num*24*365, 0, 0]]
                st_tagged[i-1] = [None, None]
        except IndexError:
            pass
        st_tagged[i] = new_tag

    def day_tag(self, st_tagged,i):
        '''
        Includes all the days of the week. It is assumed this is equivalent to
        this DAY
        '''
        self.byweekday = st_tagged[i][1]
        st_tagged[i] = [None, None]
        return

    def mth_tag(self, st_tagged,i):
        self.bymonth = st_tagged[i][1]
        st_tagged[i] = [None, None]
        return

    def rel_tag(self, st_tagged, i):
        '''
        Interprets items with the relative (rel) tag
        Words included in this tag are: yesterday (yest), today (td), tomorrow (tmrw)
        last (last), this (this), next (nxt), every (evy)
        '''
        # TODO case where word is first, second, third etc (lost as a number)
        # First find out which instance of the tag we are looking at
        word = st_tagged[i][1]
        if word == 'yest':
            #Check for phrase 'week yesterday'
            try:
                if st_tagged[i-1][1] == 'wk':
                    date = self.now + timedelta(days=6)
                    st_tagged[i-1] = [None, None]
                else:
                    date = self.now - timedelta(days=1)
                    self.dtstart = date.replace(hour=0, minute=0, second=0)
            except IndexError:
                date = self.now - timedelta(days=1)
                self.dtstart = date.replace(hour=0, minute=0, second=0)
            self.byyear = date.year
            self.bymonth = date.month
            self.bymonthday = date.day
        elif word == 'td':
            #Check for phrase 'a week today' otherwise assume today is meant
            try:
                if st_tagged[i-1][1] == 'wk':
                    date = self.now + timedelta(days=7)
                    st_tagged[i-1] = [None, None]
                else:
                    date = self.now
            except IndexError:
                date = self.now
            self.byyear = date.year
            self.bymonth = date.month
            self.bymonthday = date.day
        elif word == 'tmrw':
            #Check for phrase 'a week tomorrow' otherwise assume tomorrow is meant
            try:
                if st_tagged[i-1][1] == 'wk':
                    date = self.now + timedelta(days=8)
                    st_tagged[i-1] = [None, None]
                else:
                    date = self.now + timedelta(days=1)
            except IndexError:
                date = self.now + timedelta(days=1)
            self.byyear = date.year
            self.bymonth = date.month
            self.bymonthday = date.day
        elif word == 'last':
            try:
                last_what = st_tagged[i+1] #should be tagged as [time,day] or [day,(day)]
                if last_what[0] == 'day':
                    self.byweekday = last_what[1]
                    self.bymonthday = -1
                    st_tagged[i+1] = [None, None]
                elif last_what[1] == 'day':
                    self.bymonthday = -1
                    st_tagged[i+1] = [None, None]
                else:
                    print('Unexpected format: last')
            except IndexError:
                print('Unexpected format: last')
            st_tagged[i] = [None, None]
        elif word == 'nxt':
            # Next word should be one of: year, month, week, [month], [day]
            try:
                next_what = st_tagged[i+1]
                if next_what[0] == 'day':
                    self.dtstart = (self.now + timedelta(days=1)).replace(
                            hour=0, minute=0, second=0)

                    self.until = self.dtstart + timedelta(days=7)
                    self.byweekday = next_what[1]
                    st_tagged[i+1] = [None ,None]
                elif next_what[0] == 'mth':
                    self.bymonth = next_what[1]
                    st_tagged[i+1] = [None, None]
                elif next_what[1] == 'yr':
                    #For some reason there is no byyear thing so need to change start date
                    self.byyear = self.now.year + 1
                    self.dtstart = self.dtstart.replace(
                            day = 1, month = 1, year = self.now.year+1,
                            hour = 0, minute = 0, second = 0)

                    self.until = self.dtstart.replace(
                            day=31, month=12, hour=23, minute=59, second=59)

                    st_tagged[i+1] = [None, None]
                elif next_what[1] == 'mth':
                    date = self.now
                    if date.month < 12:
                        self.bymonth = date.month + 1
                    else:
                        self.bymonth = 1
                    st_tagged[i+1] = [None, None]
                elif next_what[1] == 'wk':
                    date = self.now.isocalendar()
                    how_many_days = 7 - date[2]  # Gets the day
                    self.dtstart = (self.now + timedelta(days=how_many_days)).replace(hour=0,minute=0,second=0)
                    self.until = self.dtstart + timedelta(days=7)
                    st_tagged[i+1] = [None, None]
                elif next_what[1] == 'day':
                    date = self.now + timedelta(day=1)
                    self.byyear = date.year
                    self.bymonth = self.month
                    self.byday = date.weekday()
                    st_tagged[i+1] = [None, None]
                else:
                    print('Unexpected format for next: unidentified word follows')
            except IndexError:
                print('Unexpected format for next: no word follows')

            st_tagged[i] = [None, None]
        elif word == 'this':
            # Next word should be one of: year, month, week, [month], [day]
            try:
                this_what = st_tagged[i+1]
                if this_what[0] == 'day':
                    self.count = 1
                    self.until = self.now + timedelta(days=7)
                    self.byweekday = this_what[1]
                    st_tagged[i+1] = [None, None]
                elif this_what[0] == 'mth':
                    self.bymonth = this_what[1]
                    st_tagged[i+1] = [None, None]
                elif this_what[1] == 'yr':
                    #For some reason there is no byyear thing so need to change end date
                    self.until = self.now.replace(day=31, month=12, hour=23, minute=59, second=59)
                    st_tagged[i+1] = [None, None]
                elif this_what[1] == 'mth':
                    date = self.now
                    self.until = date.replace(day=calendar.monthrange(self.now.year,self.now.month)[1], hour=23, minute=59, second=59)
                    st_tagged[i+1] = [None, None]
                elif this_what[1] == 'wk':
                    self.until = self.now + timedelta(days=7)
                    st_tagged[i+1] = [None, None]
                elif this_what[1] == 'day':
                    self.until = self.now.replace(hour=23, minute=59, second=59)
                    st_tagged[i+1] = [None, None]
                else:
                    print('Unexpected format for this: unidentified word follows')
            except IndexError:
                print('Unexpected format for this: no word follows')
        elif word == 'evy':
            st_tagged[i] = [None,None]
        elif word == 'in':
            #This is likely to be followed by some kind of relative time
            try:
                if st_tagged[i+1][0] == 'num':
                    #Now check word following for 'time' tag
                    if st_tagged[i+2][0] == 'time':
                        in_what = st_tagged[i+2][1]
                        #Combine into a string and check for relative time
                        time = time_finder(st_tagged[i+1][1]+in_what)
                        if time != None:
                            st_tagged[i] = time
                            st_tagged[i+1] = [None,None]
                            st_tagged[i+2] = [None,None]
                            return
                elif st_tagged[i+1][0] == 'a':
                    if st_tagged[i+2][0] == 'time':
                        in_what = st_tagged[i+2][1]
                        #Combine into a string and check for relative time
                        time = time_finder('1'+in_what)
                        if time != None:
                            st_tagged[i] = time
                            st_tagged[i+1] = [None,None]
                            st_tagged[i+2] = [None,None]
                            return
                elif st_tagged[i+1][0] == 'rel_t':
                    #Can ignore as will be picked up later
                    return
            except IndexError:
                pass
            st_tagged[i] = [None,None]
        return


    def num_tag(self,st_tagged,i):
        return

    def on_tag(self,st_tagged,i):
        return

    def len_tag(self,st_tagged,i):
        '''
        Interprets items with the 'len' tag which includes the words 'for'
        and 'till'
        '''
        word = st_tagged[i][1]
        if word == 'for':
            #TODO (sort out later)
            try:
                if st_tagged[i+1][0] == 'num':
                    n = int(st_tagged[i+1][1])
                    try:
                        if st_tagged[i+2][0] == 'time':
                            if st_tagged[i+2][1] == 'sec':
                                self.length = n
                                st_tagged[i+2] = [None,None]
                            elif st_tagged[i+2][1] == 'min':
                                self.length = n * 60
                                st_tagged[i+2] = [None,None]
                            elif st_tagged[i+2][1] == 'hr':
                                self.length = n * 3600
                                st_tagged[i+2] = [None,None]
                        else:
                            self.length = n * 3600
                    except IndexError:
                        self.length = n * 3600

                elif st_tagged[i+1][0] == 'rel_t':
                    #Convert relative time to only seconds
                    self.length = st_tagged[i+1][1][0]*3600 + st_tagged[i+1][1][1]*60 + st_tagged[i+1][1][2]
                elif st_tagged[i+1][0] == 'one':
                    #Check the unit of time the user is intending (assume hours otherwise)
                    try:
                        if st_tagged[i+2][0] == 'time':
                            if st_tagged[i+2][1] == 'sec':
                                self.length = n
                                st_tagged[i+2] = [None,None]
                            elif st_tagged[i+2][1] == 'min':
                                self.length = n * 60
                                st_tagged[i+2] = [None,None]
                            elif st_tagged[i+2][1] == 'hr':
                                self.length = n * 3600
                                st_tagged[i+2] = [None,None]
                        else:
                            #Assume hours intended
                            self.length = n * 3600
                    except IndexError:
                        self.length = n * 3600

                st_tagged[i+1] = [None,None]
            except IndexError:
                pass
        elif word == 'till':
            #TODO add functionality to consider number before and after
            st_tagged[i] = [None,None]
            try:
                if st_tagged[i+1][0] == 'num':
                    pass
                elif st_tagged[i+1][0] == 'tm':
                    pass
            except IndexError:
                pass
        st_tagged[i] = [None,None]



"""
Created on Sat Mar  4 21:17:28 2017

Converts written numbers to integers: Currenlty designed only for positive whole numbers.

@author: David
"""
#from textblob import TextBlob


d_numbers = {
        'eleven':11,
        'twelve':12,
        'thirteen':13,
        'fourteen':14,
        'fifteen':15,
        'sixteen':16,
        'seventeen':17,
        'eighteen':18,
        'nineteen':19,
        'twenty':20,
        'thirty':30,
        'fourty':40,
        'forty':40,
        'fifty':50,
        'sixty':60,
        'seventy':70,
        'eighty':80,
        'ninety':90,
        'one':1,
        'two':2,
        'three':3,
        'four':4,
        'five':5,
        'six':6,
        'seven':7,
        'eight':8,
        'nine':9,
        'ten':10,
        'hundred':100,
        'thousand':1000,
        'million':1000000,
        }

d_positions = {
        'first':1,
        'second':2,
        'third':3,
        'fourth':4,
        'fifth':5,
        'sixth':6,
        'seventh':7,
        'eighth':8,
        'ninth':9,
        'tenth':10,
        'eleventh':11,
        'twelth':12,
        'thirteenth':13,
        'fourteenth':14,
        'fifteenth':15,
        'sixteenth':16,
        'seventeenth':17,
        'eighteenth':18,
        'nineteenth':19,
        'twentieth':20,
        'thirtieth':30,
        'fourtieth':40,
        'fiftieth':50,
        'sixtieth':60,
        'seventieth':70,
        'eightieth':80,
        'ninetieth':90,
        'hundreth':100,
        'thousandth':1000,
        'millienth':1000000,
        }


def convert(s):
    '''Convert all numbers in the string to digits'''
    numbers = extract_num(s)
    #Is there any millions, thousands or hundreds?
    i = 0
    mi = -1
    th = -1
    hu = -1
    #Find indices of million, thousands and hundreds
    for num in numbers:
        if num == 1000000:
            mi = i
            th = i
            hu = i
            break
        i += 1
    i = mi if mi>-1 else 0
    for num in numbers[i:]:
        if num == 1000:
            th = i
            hu = i
            break
        i += 1
    i = th if th>-1 else 0
    for num in numbers[i:]:
        if num == 100:
            hu = i
            break
        i += 1

    total = 0
    total += convert_prefix(numbers[0:mi]) * 1000000 if mi > -1 else 0
    total += convert_prefix(numbers[mi+1:th]) * 1000 if th > -1 else 0
    total += convert_prefix(numbers[th+1:hu]) * 100 if hu > -1 else 0
    total += convert_prefix(numbers[hu+1:])

    return total

def convert_prefix(numbers):
    total = 0
    last = 0
    for num in numbers:
        if num in [100,1000]:
            #Take away the number that would have been added
            if last == 0:
                last = num
                total += last
            else:
                total -= last
                last = num*last
                total += last
        else:
            total += num
            last = num

    return total

#Read character by character from the right until something matches, store that number
#empty string and continue from that point.

def extract_num(s):
    '''
    Returns all numbers in a string as a list in the order that they appear.
    Note that it does not interpret the meaning but simply writes the numbers: e.g
    one-hundred and ninety nine will return [1,100,90,9]
    '''
    check_str = ''
    numbers = []
    for c in reversed(s):
        #Start building the string
        check_str = c + check_str
        #Check if it matches anything in the dictionary
        for num in d_numbers:
            if num in check_str:
                numbers.insert(0,d_numbers[num])
                check_str = ''
                break
        for num in d_positions:
            if num in check_str:
                numbers.insert(0,d_positions[num])
                check_str = ''
                break

    return numbers

def convert_num(s):
    '''
    Returns the input string with all numbers converted to digit format
    Currently only works for simple numbers
    '''
    s = s.split() #split up all the words
    #Check whether each word is a number in written form
    output_w = []
    previous_num = 0
    for w in s:
        check_str = ''
        numbers = []
        positional = False

        for c in reversed(w):
            check_str = c + check_str
            for num in d_numbers:
                if num in check_str:
                    numbers.insert(0,d_numbers[num])
                    check_str = ''
                    break
            for num in d_positions:
                if num in check_str:
                    numbers.insert(0,d_positions[num])
                    positional = True
                    check_str = ''
                    break
        #Try to understand the number (assumption that only two maximum will be found e.g twenty-three)
        n = 0
        if len(numbers) > 1:
            if numbers[0] > numbers[1]:
                n = numbers[0]+numbers[1]
            else:
                n = numbers[0]*numbers[1]
        elif len(numbers) == 0:
            n = w
        elif numbers[0] < previous_num:
            n += previous_num + numbers[0]
        elif previous_num == 0:
            n = numbers[0]
        else:
            n = previous_num * numbers[0]

        #Delete the previous number if this value includes it
        if previous_num > 0 and isinstance(n,int):
            del output_w[-1]

        #Add the word into the list (adjusting if it is a positional number)
        output_w.append(str(n)+'st') if positional==True else output_w.append(str(n))

        try:
            previous_num = int(n)
        except ValueError:
            previous_num = 0

    output = ' '.join(str(w) for w in output_w)
    return output


import re


_date_words = {
        'monday':['day',0],
        'mon':['day',0],
        'tuesday':['day',1],
        'tue':['day',1],
        'wednesday':['day',2],
        'wed':['day',2],
        'thursday':['day',3],
        'thur':['day',3],
        'thu':['day',3],
        'friday':['day',4],
        'fri':['day',4],
        'saturday':['day',5],
        'sat':['day',5],
        'sunday':['day',6],
        'sun':['day',6],
        'weekend':['day','wk_end'],
        'weekends':['day','wk_end'],
        'january':['mth',1],
        'jan':['mth',1],
        'february':['mth',2],
        'feb':['mth',2],
        'march':['mth',3],
        'mar':['mth',3],
        'april':['mth',4],
        'apr':['mth',4],
        'apl':['mth',4],
        'may':['mth',5],
        'june':['mth',6],
        'jun':['mth',6],
        'july':['mth',7],
        'jul':['mth',7],
        'august':['mth',8],
        'aug':['mth',8],
        'september':['mth',9],
        'sept':['mth',9],
        'sep':['mth',9],
        'october':['mth',10],
        'oct':['mth',10],
        'november':['mth',11],
        'nov':['mth',11],
        'december':['mth',12],
        'dec':['mth',12],
        }

def date_finder(st):
    '''
    Purpose of this function is to interpret numbers as they relate do dates
    returns ['date',[d,m,y]] where any of these items can be None
    Recognized formats: 
    25/04/1994, 25/04, 3rd, 25-04-1994, 25.04.1994
    
    Note: will asuume date is of the order day-month-year unless it is explicilty
    obvious e.g 04/25 is obviously mm/dd
    '''
    day = None
    month = None
    year = None
    
    #First check for the st, th, nd, rd endings
    matches = re.match(r'\d+(?=(st|th|nd|rd)$)',st)
    if matches:
        #Assumption is that a month day is referred to
        day = int(matches.group(0))
        return ['date',[day,month,year]]
        
    #Now check for the seperated values format (seperated by . or / or -)
    elif any(c in st for c in ['.','-','/','\\']):
        date_format = [None,None,None]
        nums = [None,None,None]
        
        matches = re.findall(r'\w+',st)
        #Now we need to understand each 'word' individually
        for i,match in enumerate(matches):
            #First check if it ends in st,th,nd or rd
            is_day = re.match(r'\d+(?=(st|th|nd|rd)$)',match)
            if is_day:
                 nums[i] = int(is_day.group(0))
                 date_format[i] = 'D' 
            #Now check if it is a word found in the dictionary
            tag = _date_words[match] if match in _date_words else [None,None]
            if tag[0] == 'mth':
                nums[i] = tag[1]
                date_format[i] = 'M'
            #TODO case where the user has specified a day (e.g. Mon-25th)
            #Now check for just plain numbers
            try:
                nums[i] = int(match)
                if nums[i] > 31:
                    date_format[i] = 'Y'
            except:
                pass
        return _fit(nums,date_format)
            
def _fit(nums,date_format):
    '''
    Pass date as an array of form [day,month,year]
    if position is set the number must go there
    '''
    #How many numbers are there?
    n = 0
    for num in nums:
        n += (1 if num != None else 0)
    date = [None,None,None]
    
    if n == 3:            
        #Determing if there is a specific format that must be followed        
        form = date_format
        if date_format[0] != None:
            if date_format[0] == 'D':
                form = ['D','M','Y']
            elif date_format[0] == 'M':
                form = ['M','D','Y']
            elif date_format[0] == 'Y':
                form = ['Y','M','D']
        elif date_format[1] != None:
            
            if date_format[1] == 'M':
                if date_format[2] == 'Y':
                    form = ['D','M','Y']
                elif date_format[2] == 'D':
                    form = ['Y','M','D']
            elif date_format[1] == 'D':
                form = ['M','D','Y']
        elif date_format[2] != None:
            if date_format[2] == 'D':
                form = ['Y','M','D'] 
        #Places those items that have specific designations
        for i, item in enumerate(form):
            if item == 'D':
                date[0] = nums[i]
            elif item == 'M':
                date[1] = nums[i]
            elif item == 'Y':
                date[2] = nums[i]
        #Now determine what they likely are if any other forms were given (N,M,N) or (N,N,Y)
        if form[1] == 'M':
            #must be D,M,Y or Y,M,D
            if nums[0] <= 31:
                date[0] = nums[0]
                date[2] = nums[2]
            elif nums[2] <= 31:
                date[0] = nums[2]
                date[2] = nums[0]
        elif form[2] == 'Y':
            #must be D,M,Y or M,D,Y
            if nums[1] > 12:
                date[0] = nums[1]
                date[1] = nums[0]
            elif nums[0] <= 31:
                date[0] = nums[0]
                date[1] = nums[1]
        #If no form at all is given then give the best guess
        if form[0] == form[1] == form[2] == None:
            if nums[1] > 12:
                #Must be M,D,Y
                date[0] = nums[1]
                date[1] = nums[0]
                date[2] = nums[2]
            elif nums[0] <= 31:
                date[0] = nums[0]
                date[1] = nums[1]
                date[2] = nums[2]
            else:
                date[0] = nums[2]
                date[1] = nums[1]
                date[2] = nums[0]                        
    elif n == 2:
        #Check for a specific fomat
        form = date_format
        if date_format[0] == 'D' or date_format[1] == 'M':
            form = ['D','M',None]
        elif date_format[0] == 'M' or date_format[1] == 'D':
            form = ['M','D',None]
        #Place the numbers if a format is given
        for i, item in enumerate(form):
            if item == 'D':
                date[0] = nums[i]
            elif item == 'M':
                date[1] = nums[i]
        #If there was no format specified make the best guess (assuming D,M is standard)
        if form[0] == form[1] == None:
            if nums[1] > 12:
                date[0] = nums[1]
                date[1] = nums[0]
            elif 31 >= nums[0]:
                date[0] = nums[0]
                date[1] = nums[1]
    #Check if the year is in 4-digit format
    if date[2] != None and date[2] < 1000:
        date[2] += 2000 #it is assumed we are in the current millenium
    
    return ['date',date] 

def time_finder(st):
    '''
    Recognized formats: 7am, 7:00:00 7:00am (given 'time' tag)
    7hr, 7s/sec, 7min (given 'rel_t' tag)        
    '''
    morning = None
    hour = 0
    minute = 0
    second = 0
       
    #Check for s, sec, min or hr at end of word
    matches = re.search(r'(\d+(?=(s|sec|secs)\b))',st)
    if matches:
        second = int(matches.group(1))
    matches = re.search(r'(\d+(?=(min|mins)\b))',st)
    if matches:
        minute = int(matches.group(1))
    matches = re.search(r'(\d+(?=(hr|hrs)\b))',st)
    if matches:
        hour = int(matches.group(1))
    matches = re.search(r'(\d+(?=(day)\b))',st)
    if matches:
        hour = int(matches.group(1))*24
    matches = re.search(r'(\d+(?=(wk)\b))',st)
    if matches:
        hour = int(matches.group(1))*7*24
    matches = re.search(r'(\d+(?=(mth)\b))',st)
    if matches:
        hour = int(matches.group(1))*24*30
    matches = re.search(r'(\d+(?=(yr)\b))',st)
    if matches:
        hour = int(matches.group(1))*7*24*365
      
    if hour+minute+second > 0:
        return ['rel_t',[hour,minute,second]]

    #Check for am or pm at the end of the word
    am_pm = re.search(r'(am|pm)$',st)
    try:
        if am_pm.group(1) == 'am':
            morning = True
        else:
            morning = False
    except AttributeError:
        pass    
    #Check for ##:## format
    if ':' in st or bool(re.search(r'(am|pm)$',st)):
        time = re.findall(r'(\d{1,2})',st)
        if time:
            try:
                hour = int(time[0])
                minute = int(time[1])
                second = int(time[2])
            except IndexError:
                pass
        else:
            return None
    else:
        return None
       
    if morning == False and hour < 12:
        hour += 12
    elif morning == True and hour >= 12:
        hour -= 12
        
    return ['tm',[hour,minute,second,morning]]